/*
 * Zorin Taskbar: A taskbar extension for the Zorin Desktop environment.
 * Copyright (C) 2018 Zorin OS Technologies Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Credits:
 * This file is based on code from the Dash to Dock extension by micheleg
 * and code from the Dash to Panel extension by Jason DeRose (as at
 * version 13 or commit a460d2828ac38b27b7688115910f6eb6b9f274a3)
 * Some code was also adapted from the upstream Gnome Shell source code.
 */


const Clutter = imports.gi.Clutter;
const GLib = imports.gi.GLib;
const Signals = imports.signals;
const Lang = imports.lang;
const Meta = imports.gi.Meta;
const Shell = imports.gi.Shell;
const St = imports.gi.St;
const Mainloop = imports.mainloop;

const AppDisplay = imports.ui.appDisplay;
const Dash = imports.ui.dash;
const DND = imports.ui.dnd;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;
const RemoteMenu = imports.ui.remoteMenu;
const Tweener = imports.ui.tweener;
const Util = imports.misc.util;
const Workspace = imports.ui.workspace;

const Me = imports.misc.extensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;
const WindowPreview = Me.imports.windowPreview;
const Taskbar = Me.imports.taskbar;

let DASH_ANIMATION_TIME = Dash.DASH_ANIMATION_TIME;
let DASH_ITEM_LABEL_SHOW_TIME = Dash.DASH_ITEM_LABEL_SHOW_TIME;
let DASH_ITEM_LABEL_HIDE_TIME = Dash.DASH_ITEM_LABEL_HIDE_TIME;

let LABEL_GAP = 5;
let UNGROUPED_MAX_LABEL_WIDTH = 160;
let RUNNING_INDICATOR_SIZE = 3;

let recentlyClickedAppLoopId = 0;
let recentlyClickedApp = null;
let recentlyClickedAppWindows = null;
let recentlyClickedAppIndex = 0;

let tracker = Shell.WindowTracker.get_default();

/**
 * Extend AppIcon
 *
 * - Apply a css class based on the number of windows of each application (#N);
 * - Draw a dot for each window of the application based on the default "dot" style which is hidden (#N);
 *   a class of the form "running#N" is applied to the AppWellIcon actor.
 *   like the original .running one.
 * - add a .focused style to the focused app
 * - Customize click actions.
 * - Update minimization animation target
 *
 */

var taskbarAppIcon = new Lang.Class({
    Name: 'ZorinTaskbar.TaskbarAppIcon',
    Extends: AppDisplay.AppIcon,

    _init: function(settings, appInfo, iconParams, onActivateOverride) {

        // a prefix is required to avoid conflicting with the parent class variable
        this._zorinTaskbarSettings = settings;
        this._nWindows = 0;
        this.window = appInfo.window;
        this.isLauncher = appInfo.isLauncher;

        this.parent(appInfo.app, iconParams, onActivateOverride);

        this._dot.set_width(0);
        this._focused = tracker.focus_app == this.app;
        this._isGroupApps = this._zorinTaskbarSettings.get_boolean('group-apps');

        if (appInfo.window) {
            let outbox = new St.Widget({ layout_manager: new Clutter.BinLayout() });
            let box = new St.BoxLayout();
            
            this._windowTitle = new St.Label({ 
                y_align: Clutter.ActorAlign.CENTER, 
                x_align: Clutter.ActorAlign.START, 
                style_class: 'overview-label' 
            });
            
            this._updateWindowTitle();
            this._updateWindowTitleStyle();

            this._scaleFactorChangedId = St.ThemeContext.get_for_stage(global.stage).connect('changed', () => this._updateWindowTitleStyle());

            this.actor.remove_actor(this._iconContainer);
            box.add_child(this._iconContainer);
            box.add_child(this._windowTitle);
            this.icon.actor.remove_style_class_name('overview-icon');
            box.add_style_class_name('app-well-app overview-icon');

            outbox.add_child(box);
            this.actor.set_child(outbox);
        }      

        // Monitor windows-changes instead of app state.
        // Keep using the same Id and function callback (that is extended)
        if(this._stateChangedId > 0) {
            this.app.disconnect(this._stateChangedId);
            this._stateChangedId = 0;
        }

        this._focusAppChangedId = tracker.connect('notify::focus-app', 
                                                Lang.bind(this, this._onFocusAppChanged));

        if (!this.window) {
            this._stateChangedId = this.app.connect('windows-changed',
                                                Lang.bind(this, this.onWindowsChanged));
            
            this._focusWindowChangedId = 0;
            this._titleWindowChangeId = 0;
        } else {
            this._focusWindowChangedId = global.display.connect('notify::focus-window', 
                                                Lang.bind(this, this._onFocusAppChanged));

            this._titleWindowChangeId = this.window.connect('notify::title', 
                                                Lang.bind(this, this._updateWindowTitle));
        }
        
        this._overviewWindowDragEndId = Main.overview.connect('window-drag-end',
                                                Lang.bind(this, this._onOverviewWindowDragEnd));

        this._switchWorkspaceId = global.window_manager.connect('switch-workspace',
                                                Lang.bind(this, this._onSwitchWorkspace));

        this._showDots();

        this._panelSettings = Convenience.getSettings('org.gnome.shell.extensions.zorin-panel');
        this._panelPositionId = this._panelSettings.connect('changed::panel-bottom', Lang.bind(this, this._settingsChangeRefresh));

        this._numberOverlay();

        this._signalsHandler = new Convenience.GlobalSignalsHandler();
    },

    _createWindowPreview: function() {
        // Abort if already activated
        if (this.menuManagerWindowPreview)
            return;

        // Creating a new menu manager for window previews as adding it to the
        // using the secondary menu's menu manager (which uses the "ignoreRelease"
        // function) caused the extension to crash.
        this.menuManagerWindowPreview = new PopupMenu.PopupMenuManager(this);

        this.windowPreview = new WindowPreview.thumbnailPreviewMenu(this, this.menuManagerWindowPreview);

        this.windowPreview.connect('open-state-changed', Lang.bind(this, function (menu, isPoppedUp) {
            if (!isPoppedUp)
                this._onMenuPoppedDown();
        }));
        this.menuManagerWindowPreview.addMenu(this.windowPreview);

        // grabHelper.grab() is usually called when the menu is opened. However, there seems to be a bug in the 
        // underlying gnome-shell that causes all window contents to freeze if the grab and ungrab occur
        // in quick succession in timeouts from the Mainloop (for example, clicking the icon as the preview window is opening)
        // So, instead wait until the mouse is leaving the icon (and might be moving toward the open window) to trigger the grab
        // in windowPreview.js
        let windowPreviewMenuData = this.menuManagerWindowPreview._menus[this.menuManagerWindowPreview._findMenu(this.windowPreview)];
        this.windowPreview.disconnect(windowPreviewMenuData.openStateChangeId);
        windowPreviewMenuData.openStateChangeId = this.windowPreview.connect('open-state-changed', Lang.bind(this.menuManagerWindowPreview, function(menu, open) {
            if (open) {
                if (this.activeMenu)
                    this.activeMenu.close(BoxPointer.PopupAnimation.FADE);

                // don't grab here, we are grabbing in onLeave in windowPreview.js
                //this._grabHelper.grab({ actor: menu.actor, focus: menu.sourceActor, onUngrab: Lang.bind(this, this._closeMenu, menu) });
            } else {
                this._grabHelper.ungrab({ actor: menu.actor });
            }
        }));
    },

    enableWindowPreview: function(appIcons) {
        this._createWindowPreview();

        // We first remove to ensure there are no duplicates
        this._signalsHandler.removeWithLabel('window-preview');
        this._signalsHandler.addWithLabel('window-preview', [
            this.windowPreview,
            'menu-closed',
            function(menu) {
            // enter-event doesn't fire on an app icon when the popup menu from a previously
            // hovered app icon is still open, so when a preview menu closes we need to
            // see if a new app icon is hovered and open its preview menu now.
            // also, for some reason actor doesn't report being hovered by get_hover()
            // if the hover started when a popup was opened. So, look for the actor by mouse position.
            let [x, y,] = global.get_pointer();
            let hoveredActor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, x, y);
            let appIconToOpen;
            appIcons.forEach(function (appIcon) {
                if(appIcon.actor == hoveredActor) {
                    appIconToOpen = appIcon;
                } else if(appIcon.windowPreview && appIcon.windowPreview.isOpen) {
                    appIcon.windowPreview.close();
                }
            });

            if(appIconToOpen) {
                appIconToOpen.actor.sync_hover();
                if(appIconToOpen.windowPreview && appIconToOpen.windowPreview != menu)
                    appIconToOpen.windowPreview._onEnter();
            }
            return GLib.SOURCE_REMOVE;

            }
        ]);

        this.windowPreview.enableWindowPreview();
    },

    disableWindowPreview: function() {
        this._signalsHandler.removeWithLabel('window-preview');
        if (this.windowPreview)
            this.windowPreview.disableWindowPreview();
    },

    shouldShowTooltip: function() {
        if (!this.isLauncher && getInterestingWindows(this.app).length > 0) {
            return false;
        } else {
            return this.actor.hover && (!this._menu || !this._menu.isOpen) && (!this.windowPreview || !this.windowPreview.isOpen);
        }
    },

    _onDestroy: function() {
        this.parent();
        this._destroyed = true;

        // Disconect global signals
        // stateChangedId is already handled by parent)
        if(this._focusAppChangedId)
            tracker.disconnect(this._focusAppChangedId);

        if(this._overviewWindowDragEndId)
            Main.overview.disconnect(this._overviewWindowDragEndId);

        if(this._focusWindowChangedId)
            global.display.disconnect(this._focusWindowChangedId);

        if(this._titleWindowChangeId)
            this.window.disconnect(this._titleWindowChangeId);
        
        if(this._switchWorkspaceId)
            global.window_manager.disconnect(this._switchWorkspaceId);

        if(this._scaleFactorChangedId)
            St.ThemeContext.get_for_stage(global.stage).disconnect(this._scaleFactorChangedId);

        this._panelSettings.disconnect(this._panelPositionId);
    },

    onWindowsChanged: function() {
        if (this._menu && this._menu.isOpen)
            this._menu.update();

        this._updateCounterClass();
        this.updateIconGeometry();
    },

    // Update taraget for minimization animation
    updateIconGeometry: function() {

        // If (for unknown reason) the actor is not on the stage the reported size
        // and position are random values, which might exceeds the integer range
        // resulting in an error when assigned to the a rect. This is a more like
        // a workaround to prevent flooding the system with errors.
        if (this.actor.get_stage() == null)
            return;

        let rect = new Meta.Rectangle();

        [rect.x, rect.y] = this.actor.get_transformed_position();
        [rect.width, rect.height] = this.actor.get_transformed_size();

        let windows = this.window ? [this.window] : this.app.get_windows();
        windows.forEach(function(w) {
            w.set_icon_geometry(rect);
        });
    },

    _showDots: function() {
        // Just update style if dots already exist
        if (this._focusedDots && this._unfocusedDots) {
            this._updateCounterClass();
            return;
        }

        let container = this.actor.get_children()[0];

        if (!this._isGroupApps) {
            this._focusedDots = new St.Widget({ 
                layout_manager: new Clutter.BinLayout(),
                x_expand: true, y_expand: true,
                height: this._getRunningIndicatorHeight(),
                visible: false
            });

            let mappedId = this.actor.connect('notify::mapped', () => {
                this._displayProperIndicator();
                this.actor.disconnect(mappedId);
            });
        } else {
            this._focusedDots = new St.DrawingArea({ width:1, y_expand: true });
            this._focusedDots._tweeningToWidth = null;
            this._unfocusedDots = new St.DrawingArea({width:1, y_expand: true});
            this._unfocusedDots._tweeningToWidth = null;
            
            this._focusedDots.connect('repaint', Lang.bind(this, function() {
                if(this._dashItemContainer.animatingOut) {
                    // don't draw and trigger more animations if the icon is in the middle of
                    // being added to the panel
                    return;
                }
                this._drawRunningIndicator(this._focusedDots, true);
                this._displayProperIndicator();
            }));
            
            this._unfocusedDots.connect('repaint', Lang.bind(this, function() {
                if(this._dashItemContainer.animatingOut) {
                    // don't draw and trigger more animations if the icon is in the middle of
                    // being added to the panel
                    return;
                }
                this._drawRunningIndicator(this._unfocusedDots, false);
                this._displayProperIndicator();
            }));
                
            container.add_child(this._unfocusedDots);
    
            this._updateCounterClass();
        }

        container.add_child(this._focusedDots);
    },

    _updateDotSize: function() {
        if (!this._isGroupApps) {
            this._focusedDots.height = this._getRunningIndicatorHeight();
        }

        this._settingsChangeRefresh();
    },

    _settingsChangeRefresh: function() {
        if (this._isGroupApps) {
            this._updateCounterClass();
            this._focusedDots.queue_repaint();
            this._unfocusedDots.queue_repaint();
        }

        this._displayProperIndicator(true);
    },

    _updateWindowTitleStyle: function() {
        if (this._windowTitle) {
            let maxLabelWidth = UNGROUPED_MAX_LABEL_WIDTH;
            
            this._windowTitle[(maxLabelWidth > 0 ? 'show' : 'hide')]();

            this._windowTitle.clutter_text.natural_width = maxLabelWidth;
            this._windowTitle.clutter_text.natural_width_set = true;
        }
    },

    _updateWindowTitle: function() {
        if (this._windowTitle.text != this.window.title) {
            this._windowTitle.text = this.window.title ? this.window.title : this.app.get_name();
            
            if (this._focusedDots) {
                this._displayProperIndicator();
            }
        }
    },

    popupMenu: function() {
        this._removeMenuTimeout();
        this.actor.fake_release();
        this._draggable.fakeRelease();

        if (!this._menu) {
            this._menu = new taskbarSecondaryMenu(this);
            this._menu.connect('activate-window', Lang.bind(this, function (menu, window) {
                this.activateWindow(window);
            }));
            this._menu.connect('open-state-changed', Lang.bind(this, function (menu, isPoppedUp) {
                if (!isPoppedUp)
                    this._onMenuPoppedDown();
            }));
            let id = Main.overview.connect('hiding', Lang.bind(this, function () { this._menu.close(); }));
            this._menu.actor.connect('destroy', function() {
                Main.overview.disconnect(id);
            });

            this._menuManager.addMenu(this._menu);
        }

        this.emit('menu-state-changed', true);

        if (this.windowPreview)
            this.windowPreview.close();

        this.actor.set_hover(true);
        this._menu.actor.add_style_class_name('taskbarSecondaryMenu');
        this._menu.popup();
        this._menuManager.ignoreRelease();
        this.emit('sync-tooltip');

        return false;
    },

    _onFocusAppChanged: function(windowTracker) {
        this._displayProperIndicator();
    },

    _onOverviewWindowDragEnd: function(windowTracker) {
         Mainloop.timeout_add(0, Lang.bind(this, function () {
             this._displayProperIndicator();
             return GLib.SOURCE_REMOVE;
         }));
    },

    _onSwitchWorkspace: function(windowTracker) {
        Mainloop.timeout_add(0, Lang.bind(this, function () {
             this._displayProperIndicator();
             return GLib.SOURCE_REMOVE;
         }));
    },

    _displayProperIndicator: function (force) {
        let isFocused = this._isFocusedWindow();

        if(!this._isGroupApps) {
            if (this.window && isFocused) {
                let dotPosition = Taskbar.getPosition();
                
                this._focusedDots.y_align = dotPosition == St.Side.TOP ? Clutter.ActorAlign.START : Clutter.ActorAlign.END;
                this._focusedDots.background_color = this._getRunningIndicatorColor(isFocused);
                this._focusedDots.show();
            } else if (this._focusedDots.visible) {
                this._focusedDots.hide();
            }
        } else {
            let containerWidth = this.actor.get_children()[0].width;
    
            let newFocusedDotsWidth = 0;
            let newFocusedDotsOpacity = 0;
            let newUnfocusedDotsWidth = 0;
            let newUnfocusedDotsOpacity = 0;
            
            isFocused = (tracker.focus_app == this.app);

            Mainloop.timeout_add(0, () => {
                if (!this._destroyed) {
                    if(isFocused) 
                        this.actor.add_style_class_name('focused');
                    else
                        this.actor.remove_style_class_name('focused');
                }
            });
            
            newFocusedDotsWidth = (isFocused && this._nWindows > 0) ? containerWidth : 0;
            newFocusedDotsOpacity = 255;
    
            newUnfocusedDotsWidth = containerWidth;
            newUnfocusedDotsOpacity = (!isFocused && this._nWindows > 0) ? 255 : 0;
    
            if(this._focusedDots.width != newUnfocusedDotsWidth || this._unfocusedDots.width != newFocusedDotsWidth) {
                this._animateDotDisplay(this._focusedDots, newFocusedDotsWidth, this._unfocusedDots, newUnfocusedDotsOpacity, force);
                this._animateDotDisplay(this._unfocusedDots, newUnfocusedDotsWidth, this._focusedDots, newFocusedDotsOpacity, force);
            }
        }
    },

    _animateDotDisplay: function (dots, newWidth, otherDots, newOtherOpacity, force) {
        if((dots.width != newWidth && dots._tweeningToWidth !== newWidth) || force) {
                dots._tweeningToWidth = newWidth;
                Tweener.addTween(dots,
                                { width: newWidth,
                                time: DASH_ANIMATION_TIME,
                                transition: 'easeInOutCubic',
                                onStart: Lang.bind(this, function() { 
                                    if(newOtherOpacity == 0)
                                        otherDots.opacity = newOtherOpacity;
                                }),
                                onComplete: Lang.bind(this, function() { 
                                    if(newOtherOpacity > 0)
                                        otherDots.opacity = newOtherOpacity;
                                    dots._tweeningToWidth = null;
                                })
                            });
            }
    },

    _isFocusedWindow: function() {
        let focusedWindow = global.display.focus_window;
        
        while (focusedWindow) {
            if (focusedWindow == this.window) {
                return true;
            }

            focusedWindow = focusedWindow.get_transient_for();
        }

        return false;
    },

    activate: function(button) {
        if (this.windowPreview)
            this.windowPreview.requestCloseMenu();

        let event = Clutter.get_current_event();
        let modifiers = event ? event.get_state() : 0;
        let focusedApp = tracker.focus_app;

        // Only consider SHIFT and CONTROL as modifiers (exclude SUPER, CAPS-LOCK, etc.)
        modifiers = modifiers & (Clutter.ModifierType.SHIFT_MASK | Clutter.ModifierType.CONTROL_MASK);

        // We don't change the CTRL-click behaviour: in such case we just chain
        // up the parent method and return.
        if (modifiers & Clutter.ModifierType.CONTROL_MASK) {
                // Keep default behaviour: launch new window
                // By calling the parent method I make it compatible
                // with other extensions tweaking ctrl + click
                this.parent(button);
                return;
        }

        // We check what type of click we have and if the modifier SHIFT or CONTROL is
        // being used. We then define what buttonAction should be for this event.
        let buttonAction = 0;
        if (button && button == 2 ) {
            buttonAction = "LAUNCH";
        }
        else if (button && button == 1) {
            if (modifiers & (Clutter.ModifierType.SHIFT_MASK | Clutter.ModifierType.CONTROL_MASK))
                buttonAction = "LAUNCH";
            else
                buttonAction = "MINIMIZE";
        }

        // We check if the app is running, and that the # of windows is > 0 in
        // case we use workspace isolation,
        let appIsRunning = this.app.state == Shell.AppState.RUNNING
            && getInterestingWindows(this.app).length > 0

        // We customize the action only when the application is already running
        if (appIsRunning && !this.isLauncher) {
            if (this.window) {
                //ungrouped applications behaviors
                switch (buttonAction) {
                    case 'MINIMIZE':
                        if (!Main.overview._shown && 
                            (buttonAction == 'MINIMIZE' && this._isFocusedWindow())) {
                                this.window.minimize();
                        } else {
                            Main.activateWindow(this.window);
                        }
                        
                        break;
        
                    case "LAUNCH":
                        this._launchNewInstance();
                        break;
                }
            } else {
                //grouped application behaviors
                switch (buttonAction) {
                    case "LAUNCH":
                        this._launchNewInstance();
                        break;
        
                    case "MINIMIZE":
                        // In overview just activate the app, unless the acion is explicitely
                        // requested with a keyboard modifier
                        if (!Main.overview._shown || modifiers){
                            // If we have button=2 or a modifier, allow minimization even if
                            // the app is not focused
                            if (this.app == focusedApp || button == 2 || modifiers & Clutter.ModifierType.SHIFT_MASK) {
                                // minimize all windows on double click and always in the case of primary click without
                                // additional modifiers
                                let click_count = 0;
                                if (Clutter.EventType.CLUTTER_BUTTON_PRESS)
                                    click_count = event.get_click_count();
                                let all_windows = (button == 1 && ! modifiers) || click_count > 1;
                                minimizeWindow(this.app, all_windows);
                            }
                            else
                                activateAllWindows(this.app);
                        }
                        else
                            this.app.activate();
                        break;
                }
            }
        }
        else {
            this._launchNewInstance();
        }

        Main.overview.hide();
    },

    _launchNewInstance: function() {
        if (this.app.can_open_new_window()) {
            let appActions = this.app.get_app_info().list_actions();
            let newWindowIndex = appActions.indexOf('new-window');

            this.animateLaunch();

            if (newWindowIndex < 0) {
                this.app.open_new_window(-1);
            } else {
                this.app.launch_action(appActions[newWindowIndex], global.get_current_time(), -1);
            }
        } else {
            let windows = this.window ? [this.window] : this.app.get_windows();

            if (windows.length) {
                Main.activateWindow(windows[0]);
            } else {
                this.app.activate();
            }
        }
    },

    _updateCounterClass: function() {
        let maxN = 4;
        this._nWindows = Math.min(getInterestingWindows(this.app).length, maxN);

        for (let i = 1; i <= maxN; i++){
            let className = 'running'+i;
            if(i != this._nWindows)
                this.actor.remove_style_class_name(className);
            else
                this.actor.add_style_class_name(className);
        }
    },

    _getRunningIndicatorHeight: function() {
        return RUNNING_INDICATOR_SIZE * St.ThemeContext.get_for_stage(global.stage).scale_factor;
    },

    _getRunningIndicatorColor: function() {
        let color;

        // Re-use the style - background color, and border width and color -
        // of the default dot
        let themeNode = this._dot.get_theme_node();
        color = themeNode.get_background_color();

        if(color.alpha == 0) // theme didn't provide one, use a default
            color = new Clutter.Color({ red: 128, green: 128, blue: 128, alpha: 255 });

        return color;
    },

    _drawRunningIndicator: function(area, isFocused) {
        let bodyColor = this._getRunningIndicatorColor();
        let [width, height] = area.get_surface_size();
        let cr = area.get_context();
        let n = this._nWindows;
        let size = this._getRunningIndicatorHeight();
        let padding = 0; // distance from the margin
        let yOffset = Taskbar.getPosition() == St.Side.TOP ? 0 : (height - padding -  size);
        
        if (isFocused) {
            let spacing = size; // separation between the dots
            let lineLength = width - (size*(n-1)) - (spacing*(n-1));
        
            cr.translate(0, yOffset);

            Clutter.cairo_set_source_color(cr, bodyColor);
            cr.newSubPath();
            cr.rectangle(0, 0, lineLength, size);
            for (let i = 1; i < n; i++) {
                cr.newSubPath();
                cr.rectangle(lineLength + (i*spacing) + ((i-1)*size), 0, size, size);
            }
            cr.fill();
        } else {
            // Draw the required numbers of dots
            let radius = size/2;
            let spacing = Math.ceil(width/18); // separation between the dots
        
            cr.translate((width - (2*n)*radius - (n-1)*spacing)/2, yOffset);

            Clutter.cairo_set_source_color(cr, bodyColor);
            for (let i = 0; i < n; i++) {
                cr.newSubPath();
                cr.arc((2*i+1)*radius + i*spacing, radius, radius, 0, 2*Math.PI);
            }
            cr.fill();
        }

        cr.$dispose();
    },

    _numberOverlay: function() {
        // Add label for a Hot-Key visual aid
        this._numberOverlayLabel = new St.Label();
        this._numberOverlayBin = new St.Bin({
            child: this._numberOverlayLabel,
            x_align: St.Align.START, y_align: St.Align.START,
            x_expand: true, y_expand: true
        });
        this._numberOverlayLabel.add_style_class_name('number-overlay');
        this._numberOverlayOrder = -1;
        this._numberOverlayBin.hide();

        this._iconContainer.add_child(this._numberOverlayBin);

    },

    updateNumberOverlay: function() {
        // We apply an overall scale factor that might come from a HiDPI monitor.
        // Clutter dimensions are in physical pixels, but CSS measures are in logical
        // pixels, so make sure to consider the scale.
        let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
        // Set the font size to something smaller than the whole icon so it is
        // still visible. The border radius is large to make the shape circular
        let [minWidth, natWidth] = this._iconContainer.get_preferred_width(-1);
        let font_size =  Math.round(Math.max(12, 0.3*natWidth) / scaleFactor);
        let size = Math.round(font_size*1.2);
        this._numberOverlayLabel.set_style(
           'font-size: ' + font_size + 'px;' +
           'border-radius: ' + this.icon.iconSize + 'px;' +
           'width: ' + size + 'px; height: ' + size +'px;'
        );
    },

    setNumberOverlay: function(number) {
        this._numberOverlayOrder = number;
        this._numberOverlayLabel.set_text(number.toString());
    },

    toggleNumberOverlay: function(activate) {
        if (activate && this._numberOverlayOrder > -1)
           this._numberOverlayBin.show();
        else
           this._numberOverlayBin.hide();
    },

    handleDragOver: function(source, actor, x, y, time) {
        if (source == Main.xdndHandler) {
            this.windowPreview.close();
        }
            
        return DND.DragMotionResult.CONTINUE;
    }

});

function minimizeWindow(app, param){
    // Param true make all app windows minimize
    let windows = getInterestingWindows(app);
    let current_workspace = global.screen.get_active_workspace();
    for (let i = 0; i < windows.length; i++) {
        let w = windows[i];
        if (w.get_workspace() == current_workspace && w.showing_on_its_workspace()){
            w.minimize();
            // Just minimize one window. By specification it should be the
            // focused window on the current workspace.
            if(!param)
                break;
        }
    }
}

/*
 * By default only non minimized windows are activated.
 * This activates all windows in the current workspace.
 */
function activateAllWindows(app){

    // First activate first window so workspace is switched if needed,
    // then activate all other app windows in the current workspace.
    let windows = getInterestingWindows(app);
    let w = windows[0];
    Main.activateWindow(w);
    let activeWorkspace = global.screen.get_active_workspace_index();

    if (windows.length <= 0)
        return;

    let activatedWindows = 0;

    for (let i = windows.length - 1; i >= 0; i--){
        if (windows[i].get_workspace().index() == activeWorkspace){
            Main.activateWindow(windows[i]);
            activatedWindows++;
        }
    }
}

function activateFirstWindow(app){

    let windows = getInterestingWindows(app);
    Main.activateWindow(windows[0]);
}

function cycleThroughWindows(app, reversed, shouldMinimize) {
    // Store for a little amount of time last clicked app and its windows
    // since the order changes upon window interaction
    let MEMORY_TIME=3000;

    let app_windows = getInterestingWindows(app);

    if(shouldMinimize)
        app_windows.push("MINIMIZE");

    if (recentlyClickedAppLoopId > 0)
        Mainloop.source_remove(recentlyClickedAppLoopId);
    recentlyClickedAppLoopId = Mainloop.timeout_add(MEMORY_TIME, resetRecentlyClickedApp);

    // If there isn't already a list of windows for the current app,
    // or the stored list is outdated, use the current windows list.
    if (!recentlyClickedApp ||
        recentlyClickedApp.get_id() != app.get_id() ||
        recentlyClickedAppWindows.length != app_windows.length) {
        recentlyClickedApp = app;
        recentlyClickedAppWindows = app_windows;
        recentlyClickedAppIndex = 0;
    }

    if (reversed) {
        recentlyClickedAppIndex--;
        if (recentlyClickedAppIndex < 0) recentlyClickedAppIndex = recentlyClickedAppWindows.length - 1;
    } else {
        recentlyClickedAppIndex++;
    }
    let index = recentlyClickedAppIndex % recentlyClickedAppWindows.length;
    
    if(recentlyClickedAppWindows[index] === "MINIMIZE")
        minimizeWindow(app, true);
    else
        Main.activateWindow(recentlyClickedAppWindows[index]);
}

function resetRecentlyClickedApp() {
    if (recentlyClickedAppLoopId > 0)
        Mainloop.source_remove(recentlyClickedAppLoopId);
    recentlyClickedAppLoopId=0;
    recentlyClickedApp =null;
    recentlyClickedAppWindows = null;
    recentlyClickedAppIndex = 0;

    return false;
}

function closeAllWindows(app) {
    let windows = getInterestingWindows(app);
    for (let i = 0; i < windows.length; i++)
        windows[i].delete(global.get_current_time());
}

// Filter out unnecessary windows, for instance
// nautilus desktop window.
function getInterestingWindows(app) {
    let windows = app.get_windows().filter(function(w) {
        return !w.skip_taskbar;
    });

    return windows;
}

/**
 * Extend AppIconMenu
 *
 * - set popup arrow side based on taskbar orientation
 * - Add close windows option based on quitfromdash extension
 *   (https://github.com/deuill/shell-extension-quitfromdash)
 */

var taskbarSecondaryMenu = new Lang.Class({
    Name: 'ZorinTaskbar.SecondaryMenu',
    Extends: AppDisplay.AppIconMenu,

    _init: function(source) {
        let side = Taskbar.getPosition();

        // Damm it, there has to be a proper way of doing this...
        // As I can't call the parent parent constructor (?) passing the side
        // parameter, I overwite what I need later
        this.parent(source);

        // Change the initialized side where required.
        this._arrowSide = side;
        this._boxPointer._arrowSide = side;
        this._boxPointer._userArrowSide = side;
    },

    // helper function for the quit windows abilities
    _closeWindowInstance: function(metaWindow) {
        metaWindow.delete(global.get_current_time());
    },

    _redisplay: function() {
        this.parent();

        let appMenu = this._source.app.menu;
        if(appMenu) {
            let remoteMenu = new RemoteMenu.RemoteMenu(this._source.actor, this._source.app.menu, this._source.app.action_group);
            let appMenuItems = remoteMenu._getMenuItems();
            let itemPosition = 0;
            for(let appMenuIdx in appMenuItems){
                let menuItem = appMenuItems[appMenuIdx];
                let labelText = menuItem.actor.label_actor.text;
                if(labelText == _("New Window") || labelText == _("Quit"))
                    continue;
                
                if(menuItem instanceof PopupMenu.PopupSeparatorMenuItem)
                    continue;

                // this ends up getting called multiple times, and bombing due to the signal id's being invalid
                // on a 2nd pass. disconnect the base handler and attach our own that wraps the id's in if statements
                menuItem.disconnect(menuItem._popupMenuDestroyId)
                menuItem._popupMenuDestroyId = menuItem.connect('destroy', Lang.bind(this, function(menuItem) {
                    if(menuItem._popupMenuDestroyId) {
                        menuItem.disconnect(menuItem._popupMenuDestroyId);
                        menuItem._popupMenuDestroyId = 0;
                    }
                    if(menuItem._activateId) {
                        menuItem.disconnect(menuItem._activateId);
                        menuItem._activateId = 0;
                    }
                    if(menuItem._activeChangeId) {
                        menuItem.disconnect(menuItem._activeChangeId);
                        menuItem._activeChangeId = 0;
                    }
                    if(menuItem._sensitiveChangeId) {
                        menuItem.disconnect(menuItem._sensitiveChangeId);
                        menuItem._sensitiveChangeId = 0;
                    }
                    this.disconnect(menuItem._parentSensitiveChangeId);
                    if (menuItem == this._activeMenuItem)
                        this._activeMenuItem = null;
                }));

                menuItem.actor.get_parent().remove_child(menuItem.actor);
                if(menuItem instanceof PopupMenu.PopupSubMenuMenuItem) {
                    let newSubMenuMenuItem = new PopupMenu.PopupSubMenuMenuItem(labelText);
                    let appSubMenuItems = menuItem.menu._getMenuItems();
                    for(let appSubMenuIdx in appSubMenuItems){
                        let subMenuItem = appSubMenuItems[appSubMenuIdx];
                        subMenuItem.actor.get_parent().remove_child(subMenuItem.actor);
                        newSubMenuMenuItem.menu.addMenuItem(subMenuItem);
                    }
                    this.addMenuItem(newSubMenuMenuItem, itemPosition);
                } else 
                    this.addMenuItem(menuItem, itemPosition);

                itemPosition++;
            }
            
            if(itemPosition > 0) {
                let separator = new PopupMenu.PopupSeparatorMenuItem();
                this.addMenuItem(separator, itemPosition);
            }
        }

        // quit menu
        let app = this._source.app;
        let window = this._source.window;
        let count = window ? 1 : getInterestingWindows(app).length;
        if ( count > 0) {
            this._appendSeparator();
            let quitFromTaskbarMenuText = "";
            if (count == 1)
                quitFromTaskbarMenuText = _("Quit");
            else
                quitFromTaskbarMenuText = _("Quit") + ' ' + count + ' ' + _("Windows");

            this._quitfromTaskbarMenuItem = this._appendMenuItem(quitFromTaskbarMenuText);
            this._quitfromTaskbarMenuItem.connect('activate', Lang.bind(this, function() {
                let app = this._source.app;
                let windows = window ? [window] : app.get_windows();
                for (let i = 0; i < windows.length; i++) {
                    this._closeWindowInstance(windows[i])
                }
            }));
        }
    }
});
Signals.addSignalMethods(taskbarSecondaryMenu.prototype);

/**
 * This function is used for both extendShowAppsIcon and extendDashItemContainer
 */
function ItemShowLabel()  {
    if (!this._labelText)
        return;

    this.label.set_text(this._labelText);
    this.label.opacity = 0;
    this.label.show();

    let [stageX, stageY] = this.get_transformed_position();
    let node = this.label.get_theme_node();

    let itemWidth  = this.allocation.x2 - this.allocation.x1;
    let itemHeight = this.allocation.y2 - this.allocation.y1;

    let labelWidth = this.label.get_width();
    let labelHeight = this.label.get_height();

    let x, y, xOffset, yOffset;

    let position = Taskbar.getPosition();
    let labelOffset = node.get_length('-x-offset');

    switch(position) {
      case St.Side.TOP:
          y = stageY + labelOffset + itemHeight;
          xOffset = Math.floor((itemWidth - labelWidth) / 2);
          x = stageX + xOffset;
          break;
      case St.Side.BOTTOM:
          yOffset = labelOffset;
          y = stageY - labelHeight - yOffset;
          xOffset = Math.floor((itemWidth - labelWidth) / 2);
          x = stageX + xOffset;
          break;
    }

    // keep the label inside the screen border
    // Only needed for the x coordinate.

    // Leave a few pixel gap
    let gap = LABEL_GAP;
    let monitor = Main.layoutManager.findMonitorForActor(this);
    if ( x - monitor.x < gap)
        x += monitor.x - x + labelOffset;
    else if ( x + labelWidth > monitor.x + monitor.width - gap)
        x -= x + labelWidth -( monitor.x + monitor.width) + gap;

    this.label.set_position(x, y);
    Tweener.addTween(this.label,
      { opacity: 255,
        time: DASH_ITEM_LABEL_SHOW_TIME,
        transition: 'easeOutQuad',
      });
};
