/*
 * Zorin Taskbar: A taskbar extension for the Zorin Desktop environment.
 * Copyright (C) 2018 Zorin OS Technologies Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Credits:
 * This file is based on code from the Dash to Dock extension by micheleg
 * and code from the Dash to Panel extension by Jason DeRose (as at
 * version 13 or commit a460d2828ac38b27b7688115910f6eb6b9f274a3)
 * Some code was also adapted from the upstream Gnome Shell source code.
 */

const Me = imports.misc.extensionUtils.getCurrentExtension();
const Clutter = imports.gi.Clutter;
const Convenience = Me.imports.convenience;
const Taskbar = Me.imports.taskbar;
const Lang = imports.lang;
const Main = imports.ui.main;
const Shell = imports.gi.Shell;

var zorintaskbarPanel = new Lang.Class({
    Name: 'ZorinTaskbar.Panel',

    _init: function(settings) {
        this._zorinTaskbarSettings = settings;
	//rebuild panel when taskar-position change
        this._zorinTaskbarSettings.connect('changed::taskbar-position', Lang.bind(this, function() {
            this.disable();
            this.enable();
        }));
    },

    enable : function() {
        this.panel = Main.panel;
	//choose the leftBox or the centerBox to build taskbar
        if (this._zorinTaskbarSettings.get_string('taskbar-position') == 'LEFTPANEL') {
            this.container = this.panel._leftBox;
        } else {
            this.container = this.panel._centerBox;
        }
        this.appMenu = this.panel.statusArea.appMenu;
        this.panelBox = Main.layoutManager.panelBox;

        // The main panel's connection to the "allocate" signal is competing with this extension
        // trying to move the centerBox over to the right, creating a never-ending cycle.
        // Since we don't have the ID to disconnect that handler, wrap the allocate() function 
        // it calls instead. If the call didn't originate from this file, ignore it.
        this.panel._leftBox.oldLeftBoxAllocate = this.panel._leftBox.allocate;
        this.panel._leftBox.allocate = Lang.bind(this.panel._leftBox, function(box, flags, isFromZorinTaskbar) {
            if(isFromZorinTaskbar === true) 
                this.oldLeftBoxAllocate(box, flags);
        });

        this.panel._centerBox.oldCenterBoxAllocate = this.panel._centerBox.allocate;
        this.panel._centerBox.allocate = Lang.bind(this.panel._centerBox, function(box, flags, isFromZorinTaskbar) {
            if(isFromZorinTaskbar === true) 
                this.oldCenterBoxAllocate(box, flags);
        });

        this.panel._rightBox.oldRightBoxAllocate = this.panel._rightBox.allocate;
        this.panel._rightBox.allocate = Lang.bind(this.panel._rightBox, function(box, flags, isFromZorinTaskbar) {
            if(isFromZorinTaskbar === true) 
                this.oldRightBoxAllocate(box, flags);
        });

        this._panelConnectId = this.panel.actor.connect('allocate', Lang.bind(this, function(actor,box,flags){this._allocate(actor,box,flags);}));
        this.container.remove_child(this.appMenu.container);
        this.taskbar = new Taskbar.taskbar(this._zorinTaskbarSettings);
        Main.overview.dashIconSize = this.taskbar.iconSize;

        this.container.insert_child_at_index( this.taskbar.actor, 2 );
        
        this._oldLeftBoxStyle = this.panel._leftBox.get_style();
        this._oldCenterBoxStyle = this.panel._centerBox.get_style();
        this._oldRightBoxStyle = this.panel._rightBox.get_style();

        // sync hover after a popupmenu is closed
        this.taskbar.connect('menu-closed', Lang.bind(this, function(){this.container.sync_hover();}));

        this._signalsHandler = new Convenience.GlobalSignalsHandler();
        this._signalsHandler.add(
            // Keep dragged icon consistent in size with this dash
            [
                this.taskbar,
                'icon-size-changed',
                Lang.bind(this, function() {
                    Main.overview.dashIconSize = this.taskbar.iconSize;
                })
            ],
            // This duplicate the similar signal which is in overview.js.
            // Being connected and thus executed later this effectively
            // overwrite any attempt to use the size of the default dash
            // which given the customization is usually much smaller.
            // I can't easily disconnect the original signal
            [
                Main.overview._controls.dash,
                'icon-size-changed',
                Lang.bind(this, function() {
                    Main.overview.dashIconSize = this.taskbar.iconSize;
                })
            ]
        );
    },

    disable: function () {
        this._signalsHandler.destroy();
        this.container.remove_child(this.taskbar.actor);
        this.container.add_child(this.appMenu.container);
        this.taskbar.destroy();
        this.panel.actor.disconnect(this._panelConnectId);

        // reset stored icon size  to the default dash
        Main.overview.dashIconSize = Main.overview._controls.dash.iconSize;

        this.panel._leftBox.allocate = this.panel._leftBox.oldLeftBoxAllocate;
        delete this.panel._leftBox.oldLeftBoxAllocate;

        this.panel._centerBox.allocate = this.panel._centerBox.oldCenterBoxAllocate;
        delete this.panel._centerBox.oldCenterBoxAllocate;
        
        this.panel._rightBox.allocate = this.panel._rightBox.oldRightBoxAllocate;
        delete this.panel._rightBox.oldRightBoxAllocate;

        this.appMenu = null;
        this.container = null;
        this.panel = null;
        this.taskbar = null;
        this._panelConnectId = null;
        this._signalsHandler = null;
    },

    _allocate: function(actor, box, flags) {
        let allocWidth = box.x2 - box.x1;
        let allocHeight = box.y2 - box.y1;

        let [leftMinWidth, leftNaturalWidth] = this.panel._leftBox.get_preferred_width(-1);
        let [centerMinWidth, centerNaturalWidth] = this.panel._centerBox.get_preferred_width(-1);
        let [rightMinWidth, rightNaturalWidth] = this.panel._rightBox.get_preferred_width(-1);

        let taskbarPosition = this._zorinTaskbarSettings.get_string('taskbar-position');
        let sideWidth, leftSideWidth, rightSideWidth;
        
        if (taskbarPosition == 'LEFTPANEL') {
            sideWidth = allocWidth - rightNaturalWidth - centerNaturalWidth;
        } else if (taskbarPosition == 'CENTEREDCONTENT') {
            leftSideWidth = (allocWidth - centerNaturalWidth + leftNaturalWidth - rightNaturalWidth) / 2;
            rightSideWidth = (allocWidth - centerNaturalWidth - leftNaturalWidth + rightNaturalWidth) / 2;
        } else if (taskbarPosition == 'CENTEREDMONITOR') {
            leftSideWidth = rightSideWidth = sideWidth = (allocWidth - centerNaturalWidth) / 2;
        }
        
        let childBox = new Clutter.ActorBox();
        
        childBox.y1 = 0;
        childBox.y2 = allocHeight;
        if (taskbarPosition == 'LEFTPANEL') {
            if (this.panel.actor.get_text_direction() == Clutter.TextDirection.RTL) {
                childBox.x1 = allocWidth - Math.min(Math.floor(sideWidth), leftNaturalWidth);
                childBox.x2 = allocWidth;
            } else {
                childBox.x1 = 0;
                childBox.x2 = sideWidth;
            }
        } else {
            if (this.panel.actor.get_text_direction() == Clutter.TextDirection.RTL) {
                childBox.x1 = allocWidth - leftNaturalWidth;
                childBox.x2 = allocWidth;
            } else {
                childBox.x1 = 0;
                childBox.x2 =leftNaturalWidth;
            }
        }
        this.panel._leftBox.allocate(childBox, flags, true);
        
        childBox.y1 = 0;
        childBox.y2 = allocHeight;
        if (taskbarPosition == 'LEFTPANEL') {
            if (this.panel.actor.get_text_direction() == Clutter.TextDirection.RTL) {
                childBox.x1 = rightNaturalWidth;
                childBox.x2 = childBox.x1 + centerNaturalWidth;
            } else {
                childBox.x1 = allocWidth - centerNaturalWidth - rightNaturalWidth;
                childBox.x2 = childBox.x1 + centerNaturalWidth;
            }
        } else {
            if (this.panel.actor.get_text_direction() == Clutter.TextDirection.RTL) {
                childBox.x1 = Math.max(rightNaturalWidth, rightSideWidth);
                childBox.x2 = allocWidth - Math.max(leftNaturalWidth, leftSideWidth);
            } else {
                childBox.x1 = Math.max(leftNaturalWidth, leftSideWidth);
                childBox.x2 = allocWidth - Math.max(rightNaturalWidth, rightSideWidth);
            }
        }
        this.panel._centerBox.allocate(childBox, flags, true);

        childBox.y1 = 0;
        childBox.y2 = allocHeight;
        if (this.panel.actor.get_text_direction() == Clutter.TextDirection.RTL) {
            childBox.x1 = 0;
            childBox.x2 = rightNaturalWidth;
        } else {
            childBox.x1 = allocWidth - rightNaturalWidth;
            childBox.x2 = allocWidth;
        }
        this.panel._rightBox.allocate(childBox, flags, true);

        let [cornerMinWidth, cornerWidth] = this.panel._leftCorner.actor.get_preferred_width(-1);
        let [cornerMinHeight, cornerHeight] = this.panel._leftCorner.actor.get_preferred_width(-1);
        childBox.x1 = 0;
        childBox.x2 = cornerWidth;
        childBox.y1 = allocHeight;
        childBox.y2 = allocHeight + cornerHeight;
        this.panel._leftCorner.actor.allocate(childBox, flags);

        [cornerMinWidth, cornerWidth] = this.panel._rightCorner.actor.get_preferred_width(-1);
        [cornerMinHeight, cornerHeight] = this.panel._rightCorner.actor.get_preferred_width(-1);
        childBox.x1 = allocWidth - cornerWidth;
        childBox.x2 = allocWidth;
        childBox.y1 = allocHeight;
        childBox.y2 = allocHeight + cornerHeight;
        this.panel._rightCorner.actor.allocate(childBox, flags);
    }
});
